//
//    Title: imagergba.cc
//  Created: Tue Oct  2 19:03:11 2001
//   Author: Tim Weyrich <weyrich@inf.ethz.ch>
//
// copyright (c) 2001--2003, Computer Graphics Laboratory, ETH Zuerich
//

#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <iostream.h>
#include <math.h>

#include "imagergba.h"
#include "clip.h"

#include <tiff.h>
#include <tiffio.h>

#define  HAVE_BIGENDIAN  (*(unsigned short *)"12" == 0x3132)

#define  SWAP_UINT32(x)  (( ((unsigned long)(x))                 << 24) |  \
                          ((((unsigned long)(x)) & 0x0000ff00UL) <<  8) |  \
                          ((((unsigned long)(x)) & 0x00ff0000UL) >>  8) |  \
                          ( ((unsigned long)(x))                 >> 24))

///////////////////////////////////////////////////////////////////////////////

ImageRGBA  *imageRgbaNew(int width, int height)
  /*! After use, the image must be deleted by imageRgbaFree().
   *
   *    \param width            width of the new image in pixels
   *    \param height   height of the new image in pixels
   *
   *    \return         a pointer to the new ImageRGBA struct
   */
{
  ImageRGBA  *result = NULL;
  
  assert(width >= 0);
  assert(height >= 0);

  result = (ImageRGBA *)malloc(sizeof(ImageRGBA));
  result->width = width;
  result->height = height;
  result->bytesPerPixel = 4;
  result->data = (Rgba *)malloc(width * height * sizeof(Rgba));

  return result;
}

void  imageRgbaFree(ImageRGBA *img)
  /*! The memory used by the image is freed. All pointers to the
   *  image become invalid. After use, the created image must be
   *  deleted by imageRgbaFree().
   *
   *    \param img      the image to be deleted
   */
{
  assert(img->bytesPerPixel == sizeof(Rgba));
  free(img->data);
  free(img);
}

// ----------------------------------------------------------------------------

ImageRGBA  *imageRgbaClone(const ImageRGBA *img)
  /*! The copy has got the same extent and image data.
   *
   *    \param img      source image
   *
   *    \return a pointer to the new ImageRGBA struct
   */
{
  assert(img->bytesPerPixel == sizeof(Rgba));
  
  ImageRGBA  *result = imageRgbaNew(img->width, img->height);

  if (result)
    memcpy(result->data, img->data, img->width * img->height * sizeof(Rgba));
  
  return result;
}

///////////////////////////////////////////////////////////////////////////////

void  imageRgbaClear(ImageRGBA *img)
  /*! All pixels in the image are set to white, and alpha=255. So,
   *  a call to imageRgbaClear(img) is identical to
   *  imageRgbaClear(img, 255, 255, 255, 255).
   *
   *    \param img      the image to be cleared
   */
{
  assert(img->bytesPerPixel == sizeof(Rgba));
  imageRgbaSet(img, 255, 255, 255, 255);
}

void  imageRgbaSet(ImageRGBA *img,
                   unsigned r, unsigned g, unsigned b, unsigned a)
  /*! All pixels in the image are set to (\a r, \a g, \a b, \a a).
   *
   *    \param img      the destination image
   */
{
  assert(img->bytesPerPixel == sizeof(Rgba));
  
  Rgba   val = {r, g, b, a};
  Rgba  *p, *e = img->data + img->width * img->height;

  for (p=img->data; p<e; *p++=val);
}

///////////////////////////////////////////////////////////////////////////////

ImageRGBA  *imageRgbaReadTIFF(const char *fname)
  /*! The image is read from a file named \a fname. If the file
   *  doesn't exist, or the file is no valid TIFF file, \e NULL is
   *  returned.
   *
   *    \param fname            filename
   *
   *    \return the read image; \e NULL on error
   */
{
  TIFF  *tif;
  if ((tif = TIFFOpen(fname, "r")) != NULL)
    {
      ImageRGBA  *result = NULL;
      uint32      width, height;

      TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &width);
      TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &height);
      
      if ((result = imageRgbaNew(width, height)) != NULL)
        {
          if (TIFFReadRGBAImage(tif, result->width, result->height, (unsigned long *)result->data, 0))
            { /* success -- flip image and eventually twiddle bytes
                 (watch out: I'm doing dirty tricks, assuming a
                 special memory layout of struct Rgba) */
              unsigned long  *p, *q, temp;
              int             c;
              
              if (HAVE_BIGENDIAN)
                {
                  for (p=(unsigned long *)result->data,
                         q=(unsigned long *)result->data+(result->height-1)*result->width;
                       p < q;
                       p+=result->width, q-=result->width)
                    for (c=result->width-1; c>=0; c--)
                      {
                        temp=p[c];
                        p[c]=SWAP_UINT32(q[c]);
                        q[c]=SWAP_UINT32(temp);
                      }
                  
                  if (p == q)
                    for (c=result->width-1; c>=0; c--)
                      { p[c]=SWAP_UINT32(p[c]); }
                }
              else
                {
                  unsigned long  *p, *q, temp;
                  int             c;
                  for (p=(unsigned long *)result->data,
                         q=(unsigned long *)result->data+(result->height-1)*result->width;
                       p < q;
                       p+=result->width, q-=result->width)
                    for (c=result->width-1; c>=0; c--)
                      {
                        temp=p[c];
                        p[c]=q[c];
                        q[c]=temp;
                      }
                }
            }
          else
            {
              imageRgbaFree(result);
              result = NULL;
            }
        }

      TIFFClose(tif);
      return result;
    }
  
  return NULL;
}

int  imageRgbaWriteTIFF(const char *fname, const ImageRGBA *img)
  /*! The image is written to a file named \a fname. If the file
   *  doesn't exist, a new one is created. Otherwise, the existing
   *  file is overwritten. If an error occurs, an error value !=0
   *  is returned.
   *
   *    \param fname            filename
   *
   *    \return         0 on success
   */
{
  assert(img->bytesPerPixel == sizeof(Rgba));
  
  TIFF  *tif;
  if ((tif = TIFFOpen(fname, "w")) != NULL)
    {
      int  y;

      TIFFSetField(tif, TIFFTAG_SUBFILETYPE, 0);
      TIFFSetField(tif, TIFFTAG_IMAGEWIDTH, img->width);
      TIFFSetField(tif, TIFFTAG_IMAGELENGTH, img->height);
      TIFFSetField(tif, TIFFTAG_SAMPLESPERPIXEL, 4);
      TIFFSetField(tif, TIFFTAG_BITSPERSAMPLE, 8, 8, 8, 8);
      TIFFSetField(tif, TIFFTAG_COMPRESSION, COMPRESSION_PACKBITS);
      TIFFSetField(tif, TIFFTAG_PHOTOMETRIC, PHOTOMETRIC_RGB);
      TIFFSetField(tif, TIFFTAG_DOCUMENTNAME, fname);
      TIFFSetField(tif, TIFFTAG_IMAGEDESCRIPTION, "Created With My Personal Painter Application :-)");
      TIFFSetField(tif, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
      TIFFSetField(tif, TIFFTAG_ROWSPERSTRIP, 64);
      TIFFSetField(tif, TIFFTAG_XRESOLUTION, 75.0);
      TIFFSetField(tif, TIFFTAG_YRESOLUTION, 75.0);
      TIFFSetField(tif, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
      TIFFSetField(tif, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);

      for (y=0; y<img->height; y++)
        if (TIFFWriteScanline(tif, img->data+y*img->width, y, 0) < 0)
          break;

      if (y < img->height)
        {
          cerr << "Error while writing to TIFF file." << endl;
          TIFFClose(tif);
          return 1;
        }
      else
        {
          TIFFClose(tif);
          return 0;
        }
    }
  else
    cerr << "Couldn't open `" << fname << "' for writing." << endl;
  
  return 1;
}

///////////////////////////////////////////////////////////////////////////////

void  imageRgbaDrawLine (ImageRGBA *img, int x0, int y0, int x1, int y1, Rgba colorValue)
{
  // clip line
  {
    float  x[2] = {(float)x0, (float)x1};
    float  y[2] = {(float)y0, (float)y1};

    if (clip(x, y, 0, 0, (float)img->width-1, (float)img->height-1) == 2)
      return;

    x0 = (int)floor(x[0] + 0.5);
    y0 = (int)floor(y[0] + 0.5);
    x1 = (int)floor(x[1] + 0.5);
    y1 = (int)floor(y[1] + 0.5);
  }
  
  // starting point of line
  int x = x0, y = y0;

  // direction of line
  int dx = x1-x0, dy = y1-y0;

  // increment or decrement depending on direction of line
  int sx = (dx > 0 ? 1 : (dx < 0 ? -1 : 0));
  int sy = (dy > 0 ? 1 : (dy < 0 ? -1 : 0));

  // decision parameters for voxel selection
  if ( dx < 0 ) dx = -dx;
  if ( dy < 0 ) dy = -dy;
  int ax = 2*dx, ay = 2*dy;
  int decx, decy;

  // determine largest direction component, single-step related variable
  int max = dx, var = 0;
  if ( dy > max ) { var = 1; }

  // traverse Bresenham line
  switch ( var )
    {
    case 0:  // single-step in x-direction
      for (decy=ay-dx; /**/; x += sx, decy += ay)
        {
          // process pixel
          img->data[x + y*img->width] = colorValue;

          // take Bresenham step
          if ( x == x1 ) break;
          if ( decy >= 0 ) { decy -= ax; y += sy; }
        }
      break;
    case 1:  // single-step in y-direction
      for (decx=ax-dy; /**/; y += sy, decx += ax)
        {
          // process pixel
          img->data[x + y*img->width] = colorValue;

          // take Bresenham step
          if ( y == y1 ) break;
          if ( decx >= 0 ) { decx -= ay; x += sx; }
        }
      break;
    }
}

///////////////////////////////////////////////////////////////////////////////

void  imageRgbaInvertLine(ImageRGBA *img, int x0, int y0, int x1, int y1)
{
  // clip line
  {
    float  x[2] = {(float)x0, (float)x1};
    float  y[2] = {(float)y0, (float)y1};

    if (clip(x, y, 0.0f, 0.0f, (float)(img->width-1), (float)(img->height-1)) == 2)
      return;

    x0 = (int)floor(x[0] + 0.5);
    y0 = (int)floor(y[0] + 0.5);
    x1 = (int)floor(x[1] + 0.5);
    y1 = (int)floor(y[1] + 0.5);
  }
  
  // starting point of line
  int x = x0, y = y0;

  // direction of line
  int dx = x1-x0, dy = y1-y0;

  // increment or decrement depending on direction of line
  int sx = (dx > 0 ? 1 : (dx < 0 ? -1 : 0));
  int sy = (dy > 0 ? 1 : (dy < 0 ? -1 : 0));

  // decision parameters for voxel selection
  if ( dx < 0 ) dx = -dx;
  if ( dy < 0 ) dy = -dy;
  int ax = 2*dx, ay = 2*dy;
  int decx, decy;

  // determine largest direction component, single-step related variable
  int max = dx, var = 0;
  if ( dy > max ) { var = 1; }

  // traverse Bresenham line
  switch ( var )
    {
    case 0:  // single-step in x-direction
      for (decy=ay-dx; /**/; x += sx, decy += ay)
        {
          // process pixel
          img->data[x + y*img->width].r = 255 - img->data[x + y*img->width].r;
          img->data[x + y*img->width].g = 255 - img->data[x + y*img->width].g;
          img->data[x + y*img->width].b = 255 - img->data[x + y*img->width].b;

          // take Bresenham step
          if ( x == x1 ) break;
          if ( decy >= 0 ) { decy -= ax; y += sy; }
        }
      break;
    case 1:  // single-step in y-direction
      for (decx=ax-dy; /**/; y += sy, decx += ax)
        {
          // process pixel
          img->data[x + y*img->width].r = 255 - img->data[x + y*img->width].r;
          img->data[x + y*img->width].g = 255 - img->data[x + y*img->width].g;
          img->data[x + y*img->width].b = 255 - img->data[x + y*img->width].b;

          // take Bresenham step
          if ( y == y1 ) break;
          if ( decx >= 0 ) { decx -= ay; x += sx; }
        }
      break;
    }
}

///////////////////////////////////////////////////////////////////////////////
